import numpy as np
import matplotlib.pyplot as plt
from adjustText import adjust_text
from gensim.models import Word2Vec
import tensorflow as tf
import gensim
from contextlib import contextmanager
from timeit import default_timer
import pandas as pd
import pathlib
import logging

log = logging.getLogger("TR_logger")

@contextmanager
def elapsed_timer():
    start = default_timer()
    elapser = lambda: default_timer() - start
    yield lambda: elapser()
    end = default_timer()
    elapser = lambda: end-start


def nonzero_loss(y_true, y_pred):
    y_pred_nonzero = tf.where(y_true>0,y_pred,0)
    num_labels = tf.cast(tf.math.count_nonzero(y_true),'float32')
    mae = tf.math.divide(tf.reduce_sum(tf.abs(tf.subtract(y_pred_nonzero, y_true))),num_labels)
    return mae


def get_word2vec_model(order_data, size=100, retrain=True, model_dir=pathlib.Path('./models/')):
    """
    :param data_dir: directory to find data -> later replaced by standard data format
    :param model_dir: directory to save model
    :return:
    """
    if not retrain:
        return Word2Vec.load(str(model_dir.joinpath('product2vec.model')))

    # TODO: fix on retrain
    order_data["product_id"] = train_orders["product_id"].astype(str)

    log.debug("assembling sentences...")
    sentences = order_data.groupby("order_id").apply(
        lambda order: order['product_id'].tolist())

    longest = np.max(sentences.apply(len))
    sentences = sentences.values

    log.debug("training model...")
    model = gensim.models.Word2Vec(sentences, size=size,
                                   window=longest, min_count=2, workers=4)

    model.save(model_dir+'product2vec.model')

    return model


def get_batch(vocab, model, n_batches=1):
    output = list()
    for i in range(0, n_batches):
        rand_int = np.random.randint(len(vocab), size=1)[0]
        suggestions = model.most_similar(positive=[vocab[rand_int]], topn=5)
        suggest = list()
        for i in suggestions:
            suggest.append(i[0])
        output += suggest
        output.append(vocab[rand_int])
    return output


def plot_with_labels(low_dim_embs, labels, filename='tsne.png'):
    """From Tensorflow's tutorial."""
    assert low_dim_embs.shape[0] >= len(labels), "More labels than embeddings"
    plt.figure(figsize=(10, 10))  # in inches
    texts = []
    for i, label in enumerate(labels):
        x, y = low_dim_embs[i, :]
        plt.scatter(x, y)
        texts.append(plt.text(x,y,label,size=12))
                     # xytext=(5, 2),
                     # textcoords='offset points',
                     # ha='right',
                     # va='bottom',
                     # size=15))

    #     plt.savefig(filename)
    adjust_text(texts, arrowprops=dict(arrowstyle="->", color='r', lw=0.5))
    plt.show()


def get_features(dataset, drop_categorical=True, return_df=False, return_categorical_list=False):
    """
    preprocessing on data to extract following features:

    'user_total_orders', 'user_total_items', 'total_distinct_items','user_average_basket',
    'aisle_id', 'department_id', 'product_orders', 'product_reorders',
    'product_reorder_rate', 'UP_orders', 'UP_orders_ratio',
    'UP_average_pos_in_cart', 'UP_reorder_rate', 'UP_orders_since_last',

    :param drop_categorical: if true, drop categorical features ('aisle_id' and 'department_id')
    :param return_df: if true, return features in pandas df, else return numpy array
                        also returns list of categorical features if not dropped
    """

    order_data = dataset.prior_order_df # train only on prior orders
    product_data = dataset.product_df

    log.debug('computing product features...')

    # TODO: extract reorder count and rate (or load from file)
    prods_ = pd.DataFrame()
    prods_['orders'] = order_data.groupby(order_data.product_id).size().astype(np.int32)
    product_data = product_data.join(prods_, on='product_id')

    del prods_

    log.debug('adding order info to priors...')

    ### user features

    usr_ = pd.DataFrame()
    usr_['nb_orders'] = order_data.groupby('user_id')['order_number'].max()
    users_ = pd.DataFrame()
    users_['total_items'] = order_data.groupby('user_id').size().astype(np.int16)
    users_['all_products'] = order_data.groupby('user_id')['product_id'].apply(set)
    users_['total_distinct_items'] = (users_.all_products.map(len)).astype(np.int16)
    users_ = users_.join(usr_)
    del usr_
    users_['average_basket'] = (users_.total_items / users_.nb_orders).astype(np.float32)

    ### userXproduct features

    log.debug('compute user,product features...')
    order_data['user_product'] = list(zip(order_data.product_id, order_data.user_id))

    d_ = dict()
    for row in order_data.itertuples():
        z = row.user_product
        if z not in d_:
            d_[z] = (1, row.order_number, row.add_to_cart_order)
        else:
            d_[z] = (d_[z][0] + 1, max(d_[z][1], row.order_number), d_[z][2] + row.add_to_cart_order)

    userXproduct_ = pd.DataFrame.from_dict(d_, orient='index')
    del d_

    userXproduct_.columns = ['nb_orders', 'num_orders', 'sum_pos_in_cart']
    userXproduct_.nb_orders = userXproduct_.nb_orders.astype(np.int16)
    userXproduct_.sum_pos_in_cart = userXproduct_.sum_pos_in_cart.astype(np.int16)

    log.debug(f'user X product f: {len(userXproduct_)}')

    # construct user-product list with features
    df = dataset.prior_user_prod.copy()

    log.debug('user related features')
    df['user_total_orders'] = df.user_id.map(users_.nb_orders)
    df['user_total_items'] = df.user_id.map(users_.total_items)
    df['total_distinct_items'] = df.user_id.map(users_.total_distinct_items)
    df['user_average_basket'] = df.user_id.map(users_.average_basket)

    log.debug('product related features')
    df['aisle_id'] = df.product_id.map(product_data.feature2)
    df['department_id'] = df.product_id.map(product_data.feature4)
    df['product_orders'] = df.product_id.map(product_data.orders).astype(np.int32)

    log.debug('user_X_product related features')
    df['z'] = list(zip(df.product_id, df.user_id))
    df.drop(['user_id'], axis=1, inplace=True)
    df['UP_orders'] = df.z.map(userXproduct_.nb_orders)
    df['UP_orders_ratio'] = (df.UP_orders / df.user_total_orders).astype(np.float32)
    df['UP_average_pos_in_cart'] = (df.z.map(userXproduct_.sum_pos_in_cart) / df.UP_orders).astype(np.float32)

    if log.getEffectiveLevel()<10:
        log.debug(df.dtypes)
        log.debug(df.memory_usage())

    f_to_use = ['user_total_orders', 'user_total_items', 'total_distinct_items',
                'user_average_basket', 'aisle_id', 'department_id', 'product_orders',
                'UP_orders', 'UP_orders_ratio', 'UP_average_pos_in_cart']
    df = df[f_to_use]

    categorical_features = ['aisle_id', 'department_id']

    if drop_categorical:
        df.drop(columns=categorical_features, inplace=True)

    df = df if return_df else df.to_numpy()

    if return_categorical_list:
        return df, categorical_features
    else:
        return df





"""

 For various datasets, takes input in provided format and outputs in following
 standard format (below are both pandas DataFrames):

 order_data:
    - rows: user/order/product combinations
    - columns:
        - user_id - id uniquely identifying users
        - order_number - order number for given user
        - add_to_cart_order - order product was added in given order
        - product_id - id identifying products (see below)

 product_data:
    - index : product_id numbers
    - columns:
        - feature1, feature2, ... (text/categorical features for products)
    
"""

import numpy as np
import pandas as pd
import os

IC_DATA_DIR = '../data/instacart_2017_05_01/'
SMALL_IC_DATA_DIR = '../data/instacart_2017_05_01_small/'

AWS = True
if AWS:
    import boto3
    from io import StringIO

    client = boto3.client('s3')

    bucket_name = 'bmwillett1'

    IC_DATA_DIR_AWS = 'instacart/instacart_2017_05_01_small/'


def getCSVs3(filename, ic_data_dir=IC_DATA_DIR_AWS):
    object_key = ic_data_dir + filename
    csv_obj = client.get_object(Bucket=bucket_name, Key=object_key)
    body = csv_obj['Body']
    csv_string = body.read().decode('utf-8')
    return StringIO(csv_string)

def instacart_process_AWS():
    """
    load and preprocess instacart dataset

    :param data_dir: directory containing dataset
    :return: order_data, product_data : Pandas DataFrames
    """

    # load create order_data
    orders = pd.read_csv(getCSVs3('orders.csv'), dtype={
        'order_id': np.int32,
        'user_id': np.int32,
        'order_number': np.int16},
                         usecols=['order_id', 'user_id', 'order_number'])

    priors = pd.read_csv(getCSVs3('order_products__prior.csv'), dtype={
        'order_id': np.int32,
        'product_id': np.uint16,
        'add_to_cart_order': np.int16},
                         usecols=['order_id', 'product_id', 'add_to_cart_order'])

    train = pd.read_csv(getCSVs3('order_products__train.csv'), dtype={
        'order_id': np.int32,
        'product_id': np.uint16,
        'add_to_cart_order': np.int16},
                        usecols=['order_id', 'product_id', 'add_to_cart_order'])

    orders.set_index('order_id', inplace=True)
    order_data = priors.append(train)
    order_data['user_id'] = order_data.order_id.map(orders.user_id)
    order_data['order_number'] = order_data.order_id.map(orders.order_number)
    order_data.drop(['order_id'], axis=1, inplace=True)
    order_data = order_data[['user_id', 'order_number', 'add_to_cart_order', 'product_id']]

    # load and create product_data
    products = pd.read_csv(getCSVs3('products.csv'))
    aisles = pd.read_csv(getCSVs3('aisles.csv')).set_index('aisle_id')
    departments = pd.read_csv(getCSVs3('departments.csv')).set_index('department_id')

    product_data = pd.DataFrame()
    product_data['product_id'] = products['product_id']
    product_data['feature1'] = products['product_name']
    product_data['feature2'] = products.aisle_id
    product_data['feature3'] = products.aisle_id.map(aisles.aisle)
    product_data['feature4'] = products.department_id
    product_data['feature5'] = products.department_id.map(departments.department)
    product_data.set_index('product_id', inplace=True)

    return order_data, product_data


def instacart_process(data_dir=IC_DATA_DIR):
    """
    load and preprocess instacart dataset

    :param data_dir: directory containing dataset
    :return: order_data, product_data : Pandas DataFrames
    """

    # load create order_data
    orders = pd.read_csv(data_dir + 'orders.csv', dtype={
        'order_id': np.int32,
        'user_id': np.int32,
        'order_number': np.int16},
        usecols=['order_id', 'user_id', 'order_number'])

    priors = pd.read_csv(data_dir + 'order_products__prior.csv', dtype={
        'order_id': np.int32,
        'product_id': np.uint16,
        'add_to_cart_order': np.int16},
        usecols=['order_id', 'product_id','add_to_cart_order'])

    train = pd.read_csv(data_dir + 'order_products__train.csv', dtype={
        'order_id': np.int32,
        'product_id': np.uint16,
        'add_to_cart_order': np.int16},
        usecols=['order_id', 'product_id','add_to_cart_order'])

    orders.set_index('order_id',inplace=True)
    order_data = priors.append(train)
    order_data['user_id'] = order_data.order_id.map(orders.user_id)
    order_data['order_number'] = order_data.order_id.map(orders.order_number)
    order_data.drop(['order_id'], axis=1, inplace=True)
    order_data = order_data[['user_id','order_number','add_to_cart_order','product_id']]

    # load and create product_data
    products = pd.read_csv(data_dir + 'products.csv')
    aisles = pd.read_csv(data_dir + 'aisles.csv').set_index('aisle_id')
    departments = pd.read_csv(data_dir + 'departments.csv').set_index('department_id')

    product_data = pd.DataFrame()
    product_data['product_id'] = products['product_id']
    product_data['feature1'] = products['product_name']
    product_data['feature2'] = products.aisle_id
    product_data['feature3'] = products.aisle_id.map(aisles.aisle)
    product_data['feature4'] = products.department_id
    product_data['feature5'] = products.department_id.map(departments.department)
    product_data.set_index('product_id', inplace=True)

    return order_data, product_data

def make_instacart_small(user_frac=0.01, order_frac=1, product_frac=1, data_dir=IC_DATA_DIR,
                         output_dir=SMALL_IC_DATA_DIR):
    """
    creates smaller version of instacart dataset for faster testing

    :param user_frac: fraction of users to keep
    :param order_frac: fraction of orders to keep
    :param product_frac: fraction of products to keep
    :param data_dir: input data directory
    :param output_dir: output directory for small dataset
    :return: None
    """
    priors = pd.read_csv(data_dir + 'order_products__prior.csv')
    train = pd.read_csv(data_dir + 'order_products__train.csv')
    orders = pd.read_csv(data_dir + 'orders.csv')
    products = pd.read_csv(data_dir + 'products.csv')
    aisles = pd.read_csv(data_dir + 'aisles.csv')
    departments = pd.read_csv(data_dir + 'departments.csv')

    product_ids = products.product_id.unique()
    user_ids = orders.user_id.unique()
    product_ids_small = np.random.choice(product_ids, int(len(product_ids) * product_frac), replace=False)
    user_ids_small = np.random.choice(user_ids, int(len(user_ids) * user_frac), replace=False)

    orders = orders[((orders.user_id.isin(user_ids_small)) | (user_frac == 1))]
    order_ids = orders.order_id.unique()
    order_ids_small = np.random.choice(order_ids, int(len(order_ids) * order_frac), replace=False)

    priors_small = priors[
        (priors.order_id.isin(order_ids_small)) & ((product_frac == 1) | (priors.product_id.isin(product_ids_small)))]
    train_small = train[
        (train.order_id.isin(order_ids_small)) & ((product_frac == 1) | (train.product_id.isin(product_ids_small)))]
    orders_small = orders[(orders.order_id.isin(order_ids_small))]
    products_small = products[(product_frac == 1) | (products.product_id.isin(product_ids_small))]

    priors_small.to_csv(output_dir + 'order_products__prior.csv', index=False)
    train_small.to_csv(output_dir + 'order_products__train.csv', index=False)
    orders_small.to_csv(output_dir + 'orders.csv', index=False)
    products_small.to_csv(output_dir + 'products.csv', index=False)
    aisles.to_csv(output_dir + 'aisles.csv', index=False)
    departments.to_csv(output_dir + 'departments.csv', index=False)

def train_val_split(order_data, train_frac=0.8):
    """
    splits users in order_data into train and validation sets:
        train_order_data, val_order_data

    drops users with 1 order
    for each user with >1 order, removes last order, to be predicted, produces:
        train_last_order, val_last_order

    NOTE: somewhat slow, maybe should save as csv files

    :param order_data: order_data produced by data loading function
    :param train_frac: fraction of data in train set
    :return: Pandas DataFrames: order_data_train, order_data_val, last_order_train, last_order_val
    """

    order_numbers = {}
    for row in order_data.itertuples():
        if row.user_id in order_numbers:
            order_numbers[row.user_id] = max(order_numbers[row.user_id], row.order_number)
        else:
            order_numbers[row.user_id] = row.order_number

    user_ids = order_data.user_id.unique()
    train_user_ids = np.random.choice(user_ids, int(train_frac * user_ids.shape[0]), replace=False)

    train_order_data_rows, train_last_order_rows, val_order_data_rows, val_last_order_rows = [], [], [], []
    for row in order_data.itertuples():
        if row.user_id in train_user_ids:
            if row.order_number < order_numbers[row.user_id]:
                train_order_data_rows.append(row)
            else:
                train_last_order_rows.append(row)
        else:
            if row.order_number < order_numbers[row.user_id]:
                val_order_data_rows.append(row)
            else:
                val_last_order_rows.append(row)

    train_order_data = pd.DataFrame(train_order_data_rows)
    train_last_order = pd.DataFrame(train_last_order_rows)
    val_order_data = pd.DataFrame(val_order_data_rows)
    val_last_order = pd.DataFrame(val_last_order_rows)

    return train_order_data, train_last_order, val_order_data, val_last_order


if __name__=='__main__':
    # run tests using small data set
    if not os.path.isfile(SMALL_IC_DATA_DIR+'orders.csv'):
        print("creating small dataset...")
        make_instacart_small()

    print("processing instacart data...")
    order_data, product_data = instacart_process(data_dir=SMALL_IC_DATA_DIR)
    pd.set_option('display.max_columns', None)
    print("order_data shape = ", order_data.shape)
    print(order_data.head())
    print("product_data shape = ", product_data.shape)
    print(product_data.head())

    print("doing train/val split...")
    train_order_data, train_last_order, val_order_data, val_last_order = train_val_split(order_data)
    print("train_order_data shape = ", train_order_data.shape)
    # print(train_order_data.head())
    print("train_last_order shape = ", train_last_order.shape)
    # print(train_last_order.head())
    print("val_order_data shape = ", val_order_data.shape)
    # print(val_order_data.head())
    print("val_last_order shape = ", val_last_order.shape)
    # print(val_last_order.head())



"""
Main class and methods for "Mapper-based classifier" algorithm of:

Mapper Based Classifier
Jacek Cyranka, Alexander Georges, David Meyer
arxiv:1910.08103

Thanks to Alexander Georges and Jacek Cyranka for helpful discussions, and especially to Jacek Cyranka for
sharing a Jupyter notebook implementation of the algorithm on which much of the code below is based

"""

import numpy as np
import sklearn
import os
from sklearn.neighbors import NearestNeighbors
from sklearn.pipeline import Pipeline
from gtda.mapper import FirstSimpleGap, Projection, OneDimensionalCover, make_mapper_pipeline
import pickle
from joblib import Parallel, delayed
import pathlib
import logging

log = logging.getLogger("TR_logger")

TEMP_DATA = str(pathlib.Path(__file__).parent.parent.joinpath('./temp_data/').absolute())+'/'


class MapperClassifier:
    """
    Implements topological encoding portion of mapper classifier algorithm
    To complete mapper classifier algorithm, should train output on ordinary neural network

    Implements batching using helper-class MapperClassifierBatch

    n_components: number of component PCA projection used on input data
    NRNN: number of nearest neighbors to use whenn projecting test points to clusters
    n_intervals: number of intervals used in constructing mapper graphs
    overlap_frac: overlap of intervals
    delta: parameter increasing effective size of intervals when projecting test points
    mu: parameter used in weighing train points when projecting test data
    """

    def __init__(self, n_batches=1, n_components=1, NRNN=3, remake=True, n_intervals=10, overlap_frac=0.33, delta=0.1, mu = 1e-05, label='mapper'):
        if not(type(n_batches) is int and n_batches>0):
            raise ValueError('n_batches must be a positive integer')
        self.n_batches = n_batches
        self.mappers = [MapperClassifierBatch(n_components=n_components, NRNN=NRNN, remake=remake,
                                         n_intervals=n_intervals, overlap_frac=overlap_frac,
                                         delta=delta, mu=mu, label=label+'_'+str(i)) for i in range(self.n_batches)]

    def fit(self, data):
        self.data_batches = np.array_split(data, self.n_batches)
        for mapper, data_batch in zip(self.mappers, self.data_batches):
            log.debug(f"fitting {mapper.label}...")
            mapper.fit(data_batch)

    def fit_transform(self, data):
        self.data_batches = np.array_split(data, self.n_batches)

        # fit mappers along diagonal
        diagonal = []
        for mapper, data_batch in zip(self.mappers, self.data_batches):
            log.debug(f"fitting {mapper.label}...")
            diagonal.append(mapper.fit_transform(data_batch))

        # fill rest of matrix
        output = []
        for i in range(self.n_batches):
            log.debug(f"--->transforming {i}th batch...")
            output.append([])
            for j in range(self.n_batches):
                log.debug(f"------>using {j}th mapper...")
                if i==j:
                    output[-1].append(diagonal[i])
                else:
                    output[-1].append(self.mappers[j].transform(self.data_batches[i]))
            output[-1] = np.concatenate(output[-1], axis=1)

        return np.concatenate(output)

    def transform(self, data):
        output = []
        for mapper in self.mappers:
            output.append(mapper.transform(data))

        return np.concatenate(output, axis=1)


class MapperClassifierBatch:
    """
    Same as above, but handles a single batch

    n_components: number of component PCA projection used on input data
    NRNN: number of nearest neighbors to use whenn projecting test points to clusters
    n_intervals: number of intervals used in constructing mapper graphs
    overlap_frac: overlap of intervals
    delta: parameter increasing effective size of intervals when projecting test points
    mu: parameter used in weighing train points when projecting test data
    """
    def __init__(self, n_components=1, NRNN=3, remake=True, n_intervals=10, overlap_frac=0.33, delta=0.1, mu = 1e-05, label='mapper'):
        log.debug("creating mapper-classifier...")
        self.n_components = n_components
        assert self.n_components >= 1
        self.NRNN = NRNN
        assert self.NRNN >= 1
        self.remake = remake
        self.n_intervals = n_intervals
        self.overlap_frac = overlap_frac
        self.delta = delta
        self.mu = mu
        self.label = label
        self.data = None
        self.data_len, self.data_features, self.mapper_features = None, None, None

    def fit(self, data):
        """
        constructs mapper graphs using train data and uses to transform training points to encoded space

        :param data: numpy array whose rows are training datapoints to fit to mapper-classifier
        :return: None
        """
        if len(data.shape) > 2:
            log.debug("flattening data...")
        self.data =np.reshape(data, (data.shape[0], -1))
        (self.data_len, self.data_features) = self.data.shape

        log.debug(f"fitting mapper with data of total length {self.data_len} and {self.data_features} features.")

        # get (n_components)-dimensional PCA projection of data
        log.debug("--->getting latent space rep...")
        self._getLatentRep()

        # create mapper graphs
        log.debug("--->creating mapper graphs...")
        self._runMapper()

        # create mapper graphs
        log.debug("--->pre-computing nearest neighbor classifiers...")
        self._precomputeNNs()

    def transform(self, test_data=None):
        if test_data is None:
            # uses self.data, set in fit function
            log.debug("transforming with test_data=None (transforming training data)...")

            self._makeGraphBins()
            return self.total_graphbinm

        else:
            # assign train points to graph node bins
            log.debug(f"transforming with test_data.shape = {test_data.shape}..")

            return self._test_transform(test_data)

    def fit_transform(self, data, n_batches=1):
        log.debug("fit_transform in MapperClassifier...")
        self.fit(data)
        return self.transform()

    def _getLatentRep(self):
        """
        gets latent space (PCA) representation of train data

        :return: None
        """
        if not self.remake and os.path.exists(TEMP_DATA + "%s_latent" % self.label):
            frin = open(TEMP_DATA + "%s_latent" % self.label, "rb")
            self.rep = pickle.load(frin)
            return

        log.debug("------> fitting {}-component PCA to data of shape {}...".format(self.n_components, self.data.shape))
        pca = MyPCA(n_components=self.n_components)
        pca.fit(self.data)

        self.rep = pca
        fr = open(TEMP_DATA + "%s_latent" % self.label, "wb")
        pickle.dump(self.rep, fr)
        fr.close()

    def _runMapper(self):
        """
        creates mapper graphs based on train data

        :return: None
        """
        log.debug("--->creating mappers...")
        if not self.remake and os.path.exists(TEMP_DATA + "%s_firstsimplegap_graphs" % self.label):
            fgin = open(TEMP_DATA + "%s_firstsimplegap_graphs" % self.label, "rb")
            self.graphs = pickle.load(fgin)

            fpin = open(TEMP_DATA + "%s_mapper_pipes" % self.label, "rb")
            self.mapper_pipes = pickle.load(fpin)
            return

        clusterer = FirstSimpleGap()
        self.mapper_pipes = []

        log.debug("------> creating projection components...")

        for k in range(self.n_components):
            log.debug("---------> on component {}/{}...".format(k + 1, self.n_components))
            proj = Projection(columns=k)
            filter_func = Pipeline(steps=[('pca', self.rep), ('proj', proj)])
            filtered_data = filter_func.fit_transform(self.data)
            cover = OneDimensionalCover(n_intervals=self.n_intervals, overlap_frac=self.overlap_frac, kind='balanced')
            cover.fit(filtered_data)
            mapper_pipe = make_mapper_pipeline(scaler=None,
                                               filter_func=filter_func,
                                               cover=cover,
                                               clusterer=clusterer,
                                               verbose=(log.getEffectiveLevel() == logging.DEBUG),
                                               n_jobs=1)
            mapper_pipe.set_params(filter_func__proj__columns=k)
            self.mapper_pipes.append(("PCA%d" % (k + 1), mapper_pipe))

        # try parallelization
        log.debug("------> entering parallelization...")

        self.graphs = [mapper_pipe[1].fit_transform(self.data) for mapper_pipe in self.mapper_pipes]

        #
        # self.graphs = Parallel(n_jobs=5, prefer="threads")(
        #     delayed(mapper_pipe[1].fit_transform)(self.data) for mapper_pipe in self.mapper_pipes
        # )

        fg = open(TEMP_DATA + "%s_firstsimplegap_graphs" % self.label, "wb")
        pickle.dump(self.graphs, fg)
        fg.close()

        fp = open(TEMP_DATA + "%s_mapper_pipes" % self.label, "wb")
        pickle.dump(self.mapper_pipes, fp)
        fp.close()

    def _precomputeNNs(self):
        # for each test data, translate its latent representation into interval nrs
        log.debug("--->precomputing NNs for mappers...")
        self.intervals = []
        for mapper_pipe in self.mapper_pipes:
            self.intervals.append(mapper_pipe[1].get_mapper_params()['cover'].get_fitted_intervals())

        # precompute dictionary of fitted NNeighbor for future use
        self.int_nn = {}        # = dictionary with { key = (n (= component), (i,) or  (i,i+1) = interval) :
                                # value = [knn classifier fit to points in interval, data points in interval, ids in interval] }

        for n in range(self.n_components):
            nodeids = [[] for _ in range(len(self.intervals[n]))]
            for i, x in enumerate(self.graphs[n]['node_metadata']['pullback_set_label']):
                nodeids[x] += list(self.graphs[n]['node_metadata']['node_elements'][i])
            for i in range(len(self.intervals[n])):
                if len(nodeids[i]) == 0:
                    continue
                knn = NearestNeighbors(n_neighbors=self.NRNN, metric='euclidean')
                knn.fit(self.data[nodeids[i]])
                # knn is fitted knn using data[nodeids] subset of data nodeids are also saved to further reference
                self.int_nn.update({(n, tuple([i])): [knn, self.data[nodeids[i]], np.array(nodeids[i])]})
                if i > 0:
                    knn = NearestNeighbors(n_neighbors=self.NRNN, metric='euclidean')
                    union = list(set(nodeids[i]) | set(last_nodeids))
                    knn.fit(self.data[union])
                    self.int_nn.update({(n, tuple([i - 1, i])): [knn, self.data[union], np.array(union)]})
                last_nodeids = nodeids[i]

    def _makeGraphBins(self):
        """
        transforms training points into bins according to which nodes
        of the mapper graphs they lie in.  result is a binary vector of shape
        
        (self.data_len, self.mapper_features)
        
        which is placed in self.total_graphbinm
        """
        log.debug("--->assigning train points to graph node bins...")
        
        if not self.remake and os.path.exists(TEMP_DATA + 'matrix_train_%s.csv' % self.label):
            self.total_graphbinm = np.loadtxt(TEMP_DATA + 'matrix_train_%s.csv' % self.label, delimiter=',', fmt='%f')
            return

        total_graphbin = []

        self.mapper_features = 0
        for n in range(self.n_components):
            graph = self.graphs[n]  # graph created for given projection component
            nrnodes = len(graph['node_metadata']['node_id'])  # number of nodes in graph
            graphbin = np.zeros((self.data_len, nrnodes), dtype=int)  # matrix indicating which nodes train data lies in
            for node in range(nrnodes):
                for pt in graph['node_metadata']['node_elements'][node]:
                    graphbin[pt][node] = 1

            total_graphbin.append(graphbin)
            self.mapper_features += graphbin.shape[1]

        self.total_graphbinm = np.hstack(total_graphbin)

        np.savetxt(TEMP_DATA + 'matrix_train_%s.csv' % self.label, self.total_graphbinm, delimiter=',', fmt='%f')

    def _test_transform(self, test_data_):
        """
        transforms test data points to mapper representation
        assumes fit_transform has already been called on training dataset

        :param test_data_: data to be transformed
        :return: transformed test points
        """
        log.debug("--->assigning test points to graph node bins...")

        if len(test_data_.shape) > 2:
            log.debug("flattening data...")
        test_data = np.reshape(test_data_, (test_data_.shape[0], -1))


        if not self.remake and os.path.exists(TEMP_DATA + 'matrix_test_%s.csv' % self.label):
            total_test_rep = np.loadtxt(TEMP_DATA + 'matrix_test_%s.csv' % self.label, delimiter=',', fmt='%f')
            return total_test_rep


        assert test_data.shape[1]==self.data_features
        test_data_len = test_data.shape[0]

        # project the test data
        latent_test_data = self.rep.transform(test_data)


        # for each test point compute its representation by finding its NNeighbors in train set

        def find_alphas(val, intervals, delta=self.delta):
            alphas = []
            if val < intervals[0][1]:
                return tuple([0])
            if val > intervals[-1][0]:
                return tuple([len(intervals) - 1])
            for n, intv in enumerate(intervals):
                midv = (intv[0] + intv[1]) / 2
                if abs(val - midv) < abs(intv[1] - intv[0]) * (1 + delta) / 2:
                    alphas.append(n)
            return tuple(alphas[:2])

        # for each test point,  alphas = list of intervals containing point for each component
        fullalphas = []
        for latent_testp in latent_test_data:
            alphas = []
            for n, latentv in enumerate(latent_testp):
                alphas.append(find_alphas(latentv, self.intervals[n]))
            fullalphas.append(alphas)

        test_rep = np.zeros((test_data_len, self.mapper_features))

        for i in range(test_data_len):
            nncompids = []
            nncompdists = []
            for n in range(self.n_components):
                inn = self.int_nn.get((n, fullalphas[i][n]))
                try:
                    if len(inn[2]) > self.NRNN:
                        knn = inn[0].kneighbors([test_data[i]])
                        ids = inn[2][knn[1]]
                        nncompids.append(ids.squeeze())
                        nncompdists.append(knn[0].squeeze())
                    else:
                        ids = inn[2]  # all available points
                        dists = np.linalg.norm(test_data[i] - inn[1], axis=1)
                        nncompids.append(ids)
                        nncompdists.append(dists)
                except:
                    print(inn)
                    print((n, fullalphas[i][n]))
                    print(list(self.int_nn.keys()))
                    assert False

            all_ids = np.concatenate([nncompids[j] for j in range(self.n_components)])
            all_dists = np.concatenate([nncompdists[j] for j in range(self.n_components)])
            sort_idx = np.argsort(all_dists)[:self.NRNN]
            best_ids = all_ids[sort_idx]
            best_dists = all_dists[sort_idx]

            ns = np.array(best_dists[:])
            ns = 1. / (ns + self.mu)
            ns = ns / sum(ns)
            features = ns.reshape(1, -1).dot(self.total_graphbinm[best_ids].astype(float)).squeeze()
            test_rep[i] = features

        return test_rep

class MyPCA(sklearn.decomposition.PCA):
    def fit_transform(self, X):
        return super().transform(X)

    def fit_transform(self, X, y=None):
        return super().transform(X)

class LatentRep():
    # assumes that all proj in self.projs have .fit and .transform() methods
    def __init__(self, projectors, label):
        self.projectors = projectors
        self.label = label

    def fit(self, *args, **kwargs):
        for proj in self.projectors:
            proj.fit(*args, **kwargs)

    def transform(self, X):
        rep = []
        for proj in self.projectors:
            rep.append(proj.transform(X))
        # merge all projectors into one vector
        return np.hstack(rep)


# TESTING

def run_tests():
    # set random seed for consistent tests
    np.random.seed(42)

    mapper = MapperClassifier()
    X_train = np.random.rand(200, 100)
    X_train_map = mapper.fit_transform(X_train)
    assert X_train_map.shape == (200, 285)
    X_test = np.random.rand(100, 100)
    X_test_map = mapper.transform(X_test)
    assert X_test_map.shape == (100, 285)

    log.info("mapper_class tests passed!")

if __name__ == '__main__':
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    run_tests()



"""
Definition of feature models

these extract features directly from the dataset, rather than the
latent models which train a machine learning algorithm to find features.  These are used
together with latent models to build up recommendation models
"""
import numpy as np
import lib.tools as tools
import logging

log = logging.getLogger("TR_logger")

class FeatureModel:
    """
    Base feature model other feature models inherit from
    """
    def transform(self, dataset):
        return None


class EmptyFeatureModel(FeatureModel):
    """
    Empty feature model
    """
    def transform(self, dataset):
        return np.empty(shape=(dataset.size, 0))


class MainFeatureModel(FeatureModel):
    """
    Feature model used by main recommendation models
    """
    def transform(self, dataset, drop_categorical=True, return_df=False):
        """
        extracts features from dataset and returns as numpy array

        :param dataset: dataset to extract features from
        :param drop_categorical: if true, drop categorical features
        :param return_df: if true, return result as pandas dataframe, else as numpy array of values
        :return: numpy array with features for each sample in dataset
        """
        return tools.get_features(dataset, drop_categorical=drop_categorical, return_df=return_df)


# TESTING

def run_tests(IC_DATA_DIR):
    from lib.process_data import instacart_process
    from lib.data_class import DataSet

    # set random seed for consistent tests
    np.random.seed(42)

    # load data from instacart csv files (values below use testing directory)
    order_data, product_data = instacart_process(data_dir=IC_DATA_DIR)

    # create dataset
    ic_dataset = DataSet(order_df=order_data, product_df=product_data)

    # check dataframes created correctly
    assert ic_dataset.order_df.shape == (31032, 4)
    assert ic_dataset.product_df.shape == (6126, 5)

    # create feature model and transform dataset
    feature_model = MainFeatureModel()
    X = feature_model.transform(ic_dataset)
    assert X.shape == (12214, 8)

    feature_model = MainFeatureModel()
    X = feature_model.transform(ic_dataset, return_df=True, drop_categorical=False)
    assert X.shape == (12214, 10)

    log.info("feature_models tests passed!")


if __name__ == '__main__':
    IC_DATA_DIR = '../data/instacart_2017_05_01_testing/'
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    run_tests(IC_DATA_DIR)



"""
Define Base recommendation model, all other recommendation models inherit from this
"""
import numpy as np
import logging

log = logging.getLogger("TR_logger")

class RecModel:
    """
    Base model which all main and baseline models inherit from

    Methods:

        find_threshold : pick threshold in [0,1] for prediction (ie, output of neural net > threshold => predict buy product).
                        grid search over given data set to find threshold maximizing f1 score

        evaluate : computes precision, recall, f1 score, and NCDG (normalized discountef cumulative gain) for model
                        on given data set
    """
    def __init__(self):
        self.preds = None
        self.threshold = 0.5  # default threshold, should be set by find_threshold()

    def find_threshold(self, dataset, pts=20, lower=0, upper=1):
        """
        use a give (validation) dataset to search for the threshold which gives the highest metric
        currently uses f1 score as metric to optimize

        :param dataset: validation dataset used in optimization
        :param pts: number of points in grid search of thresholds
        :param lower: lower bound of grid seacrh
        :param upper: upper bound of grid seacrh

        :return: best threshold (also set to self.threshold and used as default threshold for model)
        """
        log.debug(f"getting threshold: trying {pts} values in range ({lower},{upper})...")
        to_try = np.linspace(lower, upper, num=pts)
        best_t, best_f1 = 0, 0
        for t in to_try:
            self.evaluate(dataset, threshold=t)
            if self.metrics[3] > best_f1:  # maximize f1 score
                best_t, best_f1 = t, self.metrics[3]
        self.threshold = best_t
        return self.threshold

    def evaluate(self, dataset, threshold=None):
        """
        given dataset (with known labels assigned to dataset.labels), perform the following tests:

        [accuracy, precision, recall, f1 score, NDCG]

        place this list in variable model.metrics

        :param dataset: dataset used for testing
        :param threshold: threshold to use for testing, or if None, use self.threshold
        :return: None
        """
        log.debug("getting metrics...")
        if threshold is None:
            threshold = self.threshold

        preds = self.predict(dataset)

        prior_user_prod = dataset.prior_user_prod

        user_true = {}
        user_pred = {}
        user_pred_values = {}
        for i, row in enumerate(prior_user_prod.itertuples()):
            uid = row.user_id
            pid = row.product_id
            if uid not in user_true:
                user_true[uid], user_pred[uid], user_pred_values[uid] = [], [], {}
            user_pred_values[uid][pid] = preds[i]
            if dataset.labels[i] == 1:
                user_true[uid].append(pid)
            if preds[i] > threshold:
                user_pred[uid].append(pid)

        accs, precs, recs, f1s, ndcgs = [], [], [], [], []
        for uid in user_true:
            trues = set(user_true[uid])
            preds = set(user_pred[uid])

            tp = len(trues.intersection(preds))
            fp = len(preds) - tp
            fn = len(trues) - tp
            tn = len(user_pred_values) - tp - fp - fn

            acc = (tp+tn)/(tp+tn+fp+fn) if (tp+tn+fp+fn)>0 else 1
            prec = tp/(tp+fp) if tp+fp>0 else 1
            rec = tp/(tp+fn) if tp+fn>0 else 1
            f1 = (2*prec*rec)/(prec+rec) if prec+rec>0 else 0

            accs.append(acc)
            precs.append(prec)
            recs.append(rec)
            f1s.append(f1)

            # compute dcg
            ordered_preds = sorted((-user_pred_values[uid][pid], pid) for pid in user_pred_values[uid])
            pred_rank = {x[1]: i for i, x in enumerate(ordered_preds)}  # gives predicted rank of ith product
            ndcgs.append(sum([1/np.log(pred_rank[pid]+2) for pid in trues])/sum([1/np.log(i+2) for i in range(len(trues))])) if len(trues)>0 else 1

        self.metrics = [np.mean(accs), np.mean(precs), np.mean(recs), np.mean(f1s), np.mean(ndcgs)]
        return self.metrics



"""
A set of baseline models used as points of comparison to main models.  includes

- GetAll - predicts all products will be reordered
- Random - predicts reorders randomly (with probability set by threshold)
-

"""
import lib.tools as tools
import lightgbm as lgb
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestRegressor
from models.base_model import RecModel
import pandas as pd
import numpy as np
import pickle
import logging

log = logging.getLogger("TR_logger")


class GetAllModel(RecModel):
    """
    Baseline model that predicts all products will be reordered
    """
    def __init__(self):
        log.debug("creating getall model...")
        super().__init__()

    def fit(self, dataset):
        pass

    def predict(self, dataset):
        self.preds = np.ones(dataset.size)
        return self.preds


class RandomModel(RecModel):
    """
    Baseline model that predicts product reorders randomly
    """
    def __init__(self):
        log.debug("creating random model...")
        super().__init__()

    def fit(self, dataset):
        pass

    def predict(self, dataset):
        self.preds = np.random.rand(dataset.size)
        return self.preds


class LogisticModel(RecModel):
    def __init__(self):
        log.debug("creating logistic model...")
        super().__init__()
        self.model = LogisticRegression(random_state=0)

    def fit(self, dataset):
        log.debug("getting features...")
        X_train = tools.get_features(dataset)

        log.debug("fitting classifier...")
        self.model.fit(X_train, dataset.labels)

    def predict(self, dataset):
        log.debug("getting features...")
        X_test = tools.get_features(dataset)

        log.debug("predicting with model...")
        self.preds = self.model.predict(X_test)

        return self.preds


class RandomForestModel(RecModel):
    def __init__(self):
        log.debug("creating random forest model...")
        super().__init__()
        self.model = RandomForestRegressor(random_state=0)

    def fit(self, dataset):
        log.debug("getting features...")
        X_train = tools.get_features(dataset)

        log.debug("fitting classifier...")
        self.model.fit(X_train, dataset.labels)

    def predict(self, dataset):
        log.debug("getting features...")
        X_test = tools.get_features(dataset)

        log.debug("predicting with model...")
        self.preds = self.model.predict(X_test)

        return self.preds


class LGBoostModel(RecModel):
    def __init__(self):
        log.debug("creating lgboost model...")
        super().__init__()

    def fit(self, dataset, ROUNDS=100):
        log.debug("getting features...")
        df_train, categorical = tools.get_features(dataset, return_df=True, drop_categorical=False, return_categorical_list=True)

        self.d_train = lgb.Dataset(df_train, label=dataset.labels, categorical_feature=categorical)

        params = {
            'task': 'train',
            'boosting_type': 'gbdt',
            'objective': 'binary',
            'metric': {'binary_logloss'},
            'num_leaves': 96,
            'max_depth': 10,
            'feature_fraction': 0.9,
            'bagging_fraction': 0.95,
            'bagging_freq': 5
        }

        log.debug("fitting classifier...")
        self.model = lgb.train(params, self.d_train, ROUNDS)

    def predict(self, dataset):
        log.debug("getting features...")
        df_test, categorical = tools.get_features(dataset, return_df=True, drop_categorical=False, return_categorical_list=True)

        self.d_test = lgb.Dataset(df_test,
                                   label=dataset.labels,
                                   categorical_feature=categorical)

        log.debug("predicting with model...")
        self.preds = self.model.predict(df_test)

        return self.preds


# TESTING

def run_tests(IC_DATA_DIR):
    from lib.process_data import instacart_process
    from lib.data_class import DataSet

    # set random seed for consistent tests
    np.random.seed(42)

    # load data from instacart csv files (values below use testing directory)
    order_data, product_data = instacart_process(data_dir=IC_DATA_DIR)

    # create dataset
    ic_dataset = DataSet(order_df=order_data, product_df=product_data)

    # check dataframes created correctly
    assert ic_dataset.order_df.shape == (31032, 4)
    assert ic_dataset.product_df.shape == (6126, 5)

    # perform train-test split
    train_dataset, test_dataset = ic_dataset.train_test_split()
    assert train_dataset.order_df.shape == (24939, 4)
    assert test_dataset.order_df.shape == (6093, 4)

    # getall model
    model = GetAllModel()
    model.fit(train_dataset)
    model.predict(test_dataset)
    assert model.preds.shape == test_dataset.labels.shape

    # random model
    model = RandomModel()
    model.fit(train_dataset)
    model.predict(test_dataset)
    assert model.preds.shape == test_dataset.labels.shape

    # logistic model
    model = LogisticModel()
    model.fit(train_dataset)
    model.predict(test_dataset)
    assert model.preds.shape == test_dataset.labels.shape

    # random forest model
    model = RandomForestModel()
    model.fit(train_dataset)
    model.predict(test_dataset)
    assert model.preds.shape == test_dataset.labels.shape

    # LGBoost model
    model = LGBoostModel()
    model.fit(train_dataset)
    model.predict(test_dataset)
    assert model.preds.shape == test_dataset.labels.shape

    log.info("baseline_models tests passed!")


if __name__ == '__main__':
    IC_DATA_DIR = '../data/instacart_2017_05_01_testing/'
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    run_tests(IC_DATA_DIR)


"""
Definition of latent models

These are simple models used by RecModels to encode various features of datasets into
latent spaces, where the ideas is that data points mapping to nearby points inn the
latent spaces should be similar in some sense.  This notion of a space encoding similarity
is important for the mapper classifier algorithm to perform correctly. These are used
together with latent models to build up recommendation models
"""

import os
import numpy as np
from keras.layers import Input, Dense
from keras.models import Model
from keras.models import Model
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.decomposition import TruncatedSVD, PCA
import pathlib
import logging

import lib.tools as tools
from lib.mapper_class import MapperClassifier

log = logging.getLogger("TR_logger")
model_dir = pathlib.Path(__file__).parent.absolute()

# base class for latent models, contains common functions
class LatentModel:
    """
    Base latent model class with common methods
    should not be used directly in models

    Methods:

        fit : given training DataSet, fit latent model

        transform : given DataSet, encode points innto latent space (must call after training),
                 returns encoded output

        fit_transform : perform fit followed by transform on a DataSet
                    returns encoded output
    """
    def fit(self, dataset):
        pass

    def transform(self, dataset):
        pass

    def fit_transform(self, dataset):
        self.fit(dataset)
        return self.transform(dataset)

class EmptyLatentModel(LatentModel):
    """
    Empty latent model class
    simply returns zero-length latent vector
    Used as placeholder when a latent model is not present
    """
    def fit(self, dataset):
        pass

    def transform(self, dataset):
        return np.empty(shape=(dataset.size, 0))

# User model
class UserModel(LatentModel):
    """
    Basic user-encoding.  Uses autoencoder to encode users based on
    past purhcase history.
    """
    def __init__(self, encoding_dim=32):
        log.debug("creating UserModel...")
        self.input_dim = None
        self.encoding_dim = encoding_dim

        self.autoencoder, self.encoder, self.decoder = None, None, None

    def fit(self, dataset, epochs=50, batch_size=32):
        """
        :param dataset: input dataset used for training
        :return: None
        """
        self.input_dim = len(dataset.prod_ids)

        input = Input(shape=(self.input_dim,))

        encoded = Dense(self.encoding_dim, activation='relu')(input)
        decoded = Dense(self.input_dim, activation='sigmoid')(encoded)

        self.autoencoder = Model(input, decoded)
        self.encoder = Model(input, encoded)

        encoded_input = Input(shape=(self.encoding_dim,))
        decoder_layer = self.autoencoder.layers[-1]
        self.decoder = Model(encoded_input, decoder_layer(encoded_input))

        # TODO: way to use product latent space in encoding?
        self.autoencoder.compile(optimizer='adadelta', loss=tools.nonzero_loss)

        X_train = dataset.user_prod_matrix

        log.debug("fitting autoencoder in User Model...")
        self.autoencoder.fit(X_train, X_train,
                             epochs=epochs,
                             batch_size=batch_size,
                             shuffle=True,
                             verbose=(log.getEffectiveLevel() == logging.DEBUG))

    def transform(self, dataset):
        """
        :param dataset: input dataset used for testing
        :return: numpy array consisting of transformed points
        """
        X = dataset.user_prod_matrix
        ids = tuple([dataset.user_idx[uid] for uid in dataset.prior_user_prod.user_id])
        X_to_enc = X[ids, :]
        encoded_users = self.encoder.predict(X_to_enc)

        return encoded_users

# Topological User model
class TopUserModel(LatentModel):
    """
    Topological user latent model
    Takes previous autoencoder and includes further mapper encoding for topological protection
    """
    def __init__(self, n_components=5, NRNN=3):
        log.debug("creating TopUserModel...")
        self.user_model = UserModel()
        self.n_components, self.NRNN = n_components, NRNN
        self.mapper = MapperClassifier()
        self.X_to_map, self.X_map = None, None

    def fit(self, dataset, epochs=50, batch_size=32, retrain_mapper=True):
        log.debug("fitting TopUserModel...")
        self.user_model.fit(dataset, epochs=epochs, batch_size=batch_size)
        self.X_to_map = self.user_model.transform(dataset)

        #TODO: add logic so only trains on each user once (code written somewhere...)

        log.debug("encoding features with mapper-classifier...")
        if retrain_mapper or self.X_map is None:
            self.mapper.fit(self.X_to_map)
        log.debug(f"created mapper encoding of size {self.mapper.mapper_features}")

    def transform(self, dataset):
        log.debug("transforming TopUserModel (test data)...")
        X_to_map = self.user_model.transform(dataset)
        return self.mapper.transform(X_to_map)

    def fit_transform(self, dataset, epochs=50, batch_size=32, retrain_mapper=True):
        log.debug("fit_transform in TopUserModel...")
        self.fit(dataset, epochs=epochs, batch_size=batch_size, retrain_mapper=retrain_mapper)
        return self.mapper.transform()

# Product model
class ProductModel(LatentModel):
    """
    Basic product-encoding.  Combines two product latent models, word2vec and TFIDF,
    defined below, and projects result to n_component dimensions using PCA.
    """
    def __init__(self, n_components=10):
        self.w2v = word2vecModel()
        self.tfidf = TFIDFModel()
        self.n_components = n_components

    def fit(self, dataset):
        """
        :param dataset: input dataset used for training
        :return: None
        """
        self.w2v.fit(dataset)
        self.tfidf.fit(dataset)

        X1 = self.w2v.transform(dataset)
        X2 = self.tfidf.transform(dataset)
        X = np.hstack([X1, X2])

        self.pcamodel = PCA(n_components=self.n_components)
        self.pcamodel.fit(X)

    def transform(self, dataset):
        """
        :param dataset: input dataset used for testing
        :return: numpy array consisting of transformed points
        """
        X1 = self.w2v.transform(dataset)
        X2 = self.tfidf.transform(dataset)
        X = np.hstack([X1, X2])
        return self.pcamodel.transform(X)


class word2vecModel(LatentModel):
    def __init__(self, encoding_dim=100, retrain=False, saved_model_dir=model_dir.joinpath('saved_models')):
        self.encoding_dim, self.retrain, self.saved_model_dir = encoding_dim, retrain, saved_model_dir
        self.model = None

    def fit(self, dataset):
        """
        :param dataset: input dataset used for training
        :return: None
        """
        retrain = self.retrain or not os.path.isfile(self.saved_model_dir.joinpath('product2vec.model'))

        self.model = tools.get_word2vec_model(dataset.order_df, size=self.encoding_dim, retrain=retrain,
                                              model_dir=self.saved_model_dir)

    def transform(self, dataset):
        """
        :param dataset: input dataset used for testing
        :return: numpy array consisting of transformed points
        """
        X_enc = []
        for pid in dataset.prior_user_prod.product_id:
            if str(pid) in self.model.wv:
                X_enc.append(self.model.wv[str(pid)])
            else:
                X_enc.append(np.zeros(self.encoding_dim))

        return np.array(X_enc)


class TFIDFModel(LatentModel):
    def __init__(self, n_components=20):
        self.n_components = n_components

    def fit(self, dataset):
        """
        :param dataset: input dataset used for training
        :return: None
        """
        log.debug("extracting text data from dataset...")
        comb_frame = dataset.product_df.feature1.str.cat(" " + dataset.product_df.feature3.str.cat(" " + dataset.product_df.feature5))
        comb_frame = comb_frame.replace({"[^A-Za-z0-9 ]+": ""}, regex=True)
        comb_frame.head()
        self.vectorizer = TfidfVectorizer(stop_words='english')
        self.vectorizer.fit(comb_frame)

        log.debug("fitting TF-IDF vectorizier...")
        X = self.vectorizer.transform(comb_frame)

        log.debug("fitting SVD...")
        self.tsvd = TruncatedSVD(n_components=self.n_components)
        self.tsvd.fit(X)

    def transform(self, dataset):
        """
        :param dataset: input dataset used for testing
        :return: numpy array consisting of transformed points
        """
        df = dataset.product_df.loc[dataset.prior_user_prod.product_id]

        comb_frame = df.feature1.str.cat(" " + df.feature3.str.cat(" " + df.feature5))
        comb_frame = comb_frame.replace({"[^A-Za-z0-9 ]+": ""}, regex=True)
        comb_frame.head()
        X_test = self.vectorizer.transform(comb_frame)

        return self.tsvd.transform(X_test)



# TESTING

def run_tests(IC_DATA_DIR):
    from lib.process_data import instacart_process
    from lib.data_class import DataSet

    # set random seed for consistent tests
    np.random.seed(42)

    # load data from instacart csv files (values below use testing directory)
    order_data, product_data = instacart_process(data_dir=IC_DATA_DIR)

    # create dataset
    ic_dataset = DataSet(order_df=order_data, product_df=product_data)

    # check dataframes created correctly
    assert ic_dataset.order_df.shape == (31032, 4)
    assert ic_dataset.product_df.shape == (6126, 5)

    # perform train-test split
    train_dataset, test_dataset = ic_dataset.train_test_split()
    assert train_dataset.order_df.shape == (24939, 4)
    assert test_dataset.order_df.shape == (6093, 4)

    # create user latent model and fit to train_dataset
    user_latent = UserModel()
    user_latent.fit(train_dataset, epochs=2)
    assert user_latent.transform(train_dataset).shape == (9447, 32)

    # test encoding and decoding works as expected
    encoded_upm = user_latent.encoder.predict(train_dataset.user_prod_matrix)
    decoded_upm = user_latent.decoder.predict(encoded_upm)
    autoencoded_upm = user_latent.autoencoder.predict(train_dataset.user_prod_matrix)
    assert encoded_upm.shape == (154, 32)
    assert decoded_upm.shape == (154, 6126)
    assert (decoded_upm == autoencoded_upm).all()

    # create topological user latent model and fit to train_dataset
    top_user_latent = TopUserModel()
    top_user_latent.fit_transform(train_dataset, epochs=2)
    assert top_user_latent.transform(test_dataset).shape == (2767, 219)

    # # manually compare original UPM and autoencoder prediction
    # print(train_dataset.user_prod_matrix[:5, :6])
    # print(autoencoded_upm[:5, :6])

    # transform test_dataset
    assert user_latent.transform(test_dataset).shape == (2767, 32)

    # create word2vec latent model and fit to train_dataset and transform test_dataset
    # TODO: not retraining correctly...
    w2v_model = word2vecModel()
    w2v_model.fit(train_dataset)
    log.debug(w2v_model.model)
    assert w2v_model.transform(test_dataset).shape == (2767, 100)

    # create TFIDF latent model and fit to train_dataset and transform test_dataset
    tfidf_model = TFIDFModel()
    tfidf_model.fit(train_dataset)
    assert tfidf_model.transform(test_dataset).shape == (2767, 20)

    # create product latent model (combination of previous two) and fit to train_dataset and transform test_dataset
    product_latent = ProductModel()
    product_latent.fit(train_dataset)
    assert product_latent.transform(test_dataset).shape == (2767, 10)

    log.info("latent_models tests passed!")

if __name__ == '__main__':
    IC_DATA_DIR = '../data/instacart_2017_05_01_testing/'
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    model = run_tests(IC_DATA_DIR)



"""
Definition of standard DataSet class, used by all models
"""
import numpy as np
import pandas as pd
import logging

log = logging.getLogger("TR_logger")

class DataSet:
    """
    Dataset class

    methods:

        train_test_split : splits dataset into two new datasets for fitting and training

        make_adversarial : returns dataset with num_switches of each user's products replaced by a random product

        _prior_last_split : splits order history for each user into last order and all others, for predicting final order
                            saves in variable prior_order_df.  also creates labels and size

        _make_user_prod_matrix : makes user product matrix, used for autoencoder
                                saves in user_prod_matrix

    variables:

        user_ids: numpy array of user ids
        prod_ids: numpy array of product ids

        prior_order_df: dataframe consisting of all but last order for each user
        prior_user_prod: dataframe with user-product pairs from prior orders
        labels: binary vector keeping track of which products were purchased on last order, target for prediction
        size: number of data points.  depends on specific targets, but here we take it to be size of user_prod_pairs

        user_prod_matrix : matrix of user and product interactions
    """

    def __init__(self, order_df=None, product_df=None, user_df=None):
        log.debug("creating new dataset...")
        self.order_df, self.product_df, self.user_df = None, None, None
        if order_df is not None:
            self._add_order_data(order_df)
        if product_df is not None:
            self._add_product_data(product_df)
        if user_df is not None:
            self._add_user_data(user_df)

        # set when called by _prior_last_split()
        self._prior_order_df, self._prior_user_prod, self._labels, self._size = None, None, None, None

        # set when called by _make_user_prod_matrix()
        self._user_prod_matrix = None

    def _add_order_data(self, order_df):
        log.debug("adding order dataframe...")
        self.order_df = order_df
        self.user_ids = order_df.user_id.unique()
        self.prod_ids = order_df.product_id.unique()
        self.user_idx = {uid: i for i, uid in enumerate(self.user_ids)}
        self.prod_idx = {pid: i for i, pid in enumerate(self.prod_ids)}

    def _add_product_data(self, product_df):
        log.debug("adding product dataframe...")
        self.product_df = product_df

    def _add_user_data(self, user_df):
        log.debug("adding user dataframe...")
        self.user_df = user_df

    def train_test_split(self, test_frac=0.25, inherit_products=True, shuffle=True):
        """
         returns two Dataset objects, train and test, split into fractions (1-test_size) and test_size of the users
        if inherit_products, preserves list of all products from parent

        :param test_size: fraction of users to place in test dataset
        :param inherit_products: whether to inherit list of products from original dataset
        :return: test and train DataSet objects
        """
        log.debug(f"train/test split with test_frac={test_frac}...")
        user_ids = np.copy(self.user_ids)
        if shuffle:
            np.random.shuffle(user_ids)

        # number of train samples
        num_train = int((1-test_frac)*len(user_ids))

        # split user_ids among train and test sets
        train_user_ids, test_user_ids = user_ids[:num_train], user_ids[num_train:]

        # create train and test dataframes
        train_order_df = self.order_df[self.order_df.user_id.isin(train_user_ids)].copy()
        test_order_df = self.order_df[self.order_df.user_id.isin(test_user_ids)].copy()

        # create train and test DataSets
        train_dataset = DataSet(order_df=train_order_df, product_df=self.product_df)
        test_dataset = DataSet(order_df=test_order_df, product_df=self.product_df)

        # if inherit_products, replace prod_ids and prod_idx by those of parent
        if inherit_products:
            train_dataset.prod_ids = self.prod_ids.copy()
            train_dataset.prod_idx = self.prod_idx.copy()

            test_dataset.prod_ids = self.prod_ids.copy()
            test_dataset.prod_idx = self.prod_idx.copy()

        return train_dataset, test_dataset


    def make_adversarial(self, num_switches=1, inherit_ids=True):
        """
        returns dataset with num_switches of each user's prior products (all but last order) replaced by a random product
        used for testing robustness of models
        
        :param num_switches: number of random switches (with replacement) to perform for each user
        :param inherit_ids: if true, inherits user_ids and prod_ids from original
        :return: new Dataset object
        """
        log.debug(f"making adversarial dataset with {num_switches} switches...")

        idxs = {user_id : [] for user_id in self.user_ids}

        for i, row in enumerate(self.order_df.itertuples()):
            idxs[row.user_id].append(i)

        to_swap = []
        for uid in self.user_ids:
            to_swap += list(np.random.choice(idxs[uid], size=num_switches))

        all_pids = self.order_df.product_id.unique() # do not use self.prod_ids as this can be larger if inherited
        swapped_pids = np.random.choice(all_pids, size=len(to_swap))

        new_pids = np.copy(self.order_df.product_id.values)

        to_swap = sorted(to_swap)

        for i, pid in zip(to_swap, swapped_pids):
            new_pids[i] = pid

        new_order_df = self.order_df.copy()
        new_order_df.product_id = new_pids

        new_dataset = DataSet(order_df=new_order_df, product_df=self.product_df)

        if inherit_ids:
            new_dataset.user_ids = self.user_ids.copy()
            new_dataset.user_idx = self.user_idx.copy()
            new_dataset.prod_ids = self.prod_ids.copy()
            new_dataset.prod_idx = self.prod_idx.copy()

        return new_dataset

    @property
    def prior_order_df(self):
        if self._prior_order_df is None:
            self._prior_last_split()
        return self._prior_order_df

    @property
    def prior_user_prod(self):
        if self._prior_user_prod is None:
            self._prior_last_split()
        return self._prior_user_prod

    @property
    def labels(self):
        if self._labels is None:
            self._prior_last_split()
        return self._labels

    @property
    def size(self):
        if self._size is None:
            self._prior_last_split()
        return self._size

    def _prior_last_split(self):
        """
        Splits each users orders into all but last and last orders
        creates new dataframe prior_order_df, as well as labels and size

        :return: None
        """
        log.debug("performing split into prior and last orders...")

        # TODO: improve using groupby to get max order_number for each user
        log.debug("finding number of orders for each user...")
        order_numbers = {}
        for row in self.order_df.itertuples():
            if row.user_id in order_numbers:
                order_numbers[row.user_id] = max(order_numbers[row.user_id], row.order_number)
            else:
                order_numbers[row.user_id] = row.order_number

        log.debug("creating rows of new dataframes...")
        prior_order_df_rows, last_order_df_rows = [], []
        for row in self.order_df.itertuples():
            if row.order_number < order_numbers[row.user_id]:
                prior_order_df_rows.append(row)
            else:
                last_order_df_rows.append(row)

        self._prior_order_df = pd.DataFrame(prior_order_df_rows).drop(columns=['Index'])
        self._prior_user_prod = self._prior_order_df[['user_id', 'product_id']].drop_duplicates().reset_index(drop=True)
        last_order_df = pd.DataFrame(last_order_df_rows).set_index(['user_id', 'product_id'])

        log.debug("creating labels...")
        self._labels = []
        for row in self._prior_user_prod.itertuples():
            self._labels.append(int((row.user_id, row.product_id) in last_order_df.index))
        self._labels = np.array(self._labels)

        self._size = np.shape(self._labels)[0]

    @property
    def user_prod_matrix(self):
        if self._user_prod_matrix is None:
            self._make_user_prod_matrix()
        return self._user_prod_matrix

    def _make_user_prod_matrix(self):
        """
        makes matrix of user and product purchase history, saves in self._user_prod_matrix
        used for user autoencoder

        :return: None
        """
        if self.order_df is None:
            log.error("Order dataframe not defined!")
            return

        log.debug("creating user_prod_matrix...")

        updf = self.order_df.groupby(["user_id", "product_id"])["product_id"].count().reset_index(name="count")

        user_product_dict = {}
        for row in updf.itertuples():
            uid, pid, cnt = row.user_id, row.product_id, row.count
            if row.user_id not in user_product_dict:
                user_product_dict[uid] = {}
            user_product_dict[uid][pid] = cnt

        user_total = {}
        for uid in self.user_ids:
            user_total[uid] = sum([user_product_dict[uid][pid] for pid in user_product_dict[uid].keys()])

        self._user_prod_matrix = np.zeros((len(self.user_ids), len(self.prod_ids)))
        for uid in self.user_ids:
            for pid in user_product_dict[uid]:
                self.user_prod_matrix[self.user_idx[uid], self.prod_idx[pid]] = user_product_dict[uid][pid] / user_total[uid]

        log.debug(f"created user_prod_matrix of shape {self._user_prod_matrix.shape}")


# TESTING

def run_tests(IC_DATA_DIR):
    from lib.process_data import instacart_process

    # set random seed for consistent tests
    np.random.seed(42)

    # load data from instacart csv files (values below use testing directory)
    order_data, product_data = instacart_process(data_dir=IC_DATA_DIR)

    # create dataset
    ic_dataset = DataSet(order_df=order_data, product_df=product_data)

    # check dataframes created correctly
    assert ic_dataset.order_df.shape == (31032, 4)
    assert ic_dataset.product_df.shape == (6126, 5)

    # check user and product ids created correctly
    assert ic_dataset.user_ids.shape == (206,)
    assert len(ic_dataset.user_idx) == 206
    assert ic_dataset.prod_ids.shape == (6126,)
    assert len(ic_dataset.prod_idx) == 6126

    # perform train-test split
    train_dataset, test_dataset = ic_dataset.train_test_split()
    assert train_dataset.order_df.shape == (24939, 4)
    assert train_dataset.product_df.shape == (6126, 5)
    assert test_dataset.order_df.shape == (6093, 4)
    assert test_dataset.product_df.shape == (6126, 5)

    # check that prod_ids inherited correctly
    assert (train_dataset.prod_ids == ic_dataset.prod_ids).all()
    assert (test_dataset.prod_ids == ic_dataset.prod_ids).all()

    # create adversarial dataset
    adv_dataset = ic_dataset.make_adversarial()
    assert adv_dataset.order_df.shape == (31032, 4)
    assert adv_dataset.product_df.shape == (6126, 5)
    assert (adv_dataset.user_ids == ic_dataset.user_ids).all()

    # number of places old and new dfs differ, should be number of users (unless swap product for self; very unlikely)
    assert np.sum(ic_dataset.order_df.product_id.values != adv_dataset.order_df.product_id.values) == 206

    # test prior-last order split
    assert ic_dataset.prior_order_df.shape == (28906, 4)
    assert ic_dataset.prior_user_prod.shape == (12214, 2)
    assert ic_dataset.labels.shape == (12214,)
    assert ic_dataset.size == 12214

    # test user-product matrix
    assert ic_dataset.user_prod_matrix.shape == (206, 6126)

    log.info("data_class tests passed!")

if __name__ == '__main__':
    IC_DATA_DIR = '../data/instacart_2017_05_01_testing/'
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    run_tests(IC_DATA_DIR)



"""
Definition of main recommendation models

Summary:

    NonTopModel : TBA

    TopModel : TBA

"""
import numpy as np
import tensorflow as tf
from keras.layers import Input, Dense
from keras.models import Model
import logging

from models.base_model import RecModel
from models.feature_models import EmptyFeatureModel
from models.latent_models import EmptyLatentModel
from lib.mapper_class import MapperClassifier

log = logging.getLogger("TR_logger")

class NonTopModel(RecModel):
    """
    Main non-topological recommendation model

    details TBA
    """
    def __init__(self, user_latent_model=None, product_latent_model=None, feature_model=None,
                 h_dims = (100,50), output_dim=1):
        super().__init__()
        log.debug("creating main non-topological model...")
        if user_latent_model is None:
            user_latent_model = EmptyLatentModel()
        if product_latent_model is None:
            product_latent_model = EmptyLatentModel()
        if feature_model is None:
            feature_model = EmptyFeatureModel()
        self.user_latent_model, self.product_latent_model, self.feature_model = user_latent_model, product_latent_model, feature_model

        self.input_dim, self.h_dims, self.output_dim = None, h_dims, output_dim
        self.final_model = None

    def fit(self, dataset, fit_latent=True, epochs=50):
        """
        Fits model to training dataset

        :param dataset: dataset used for training
        :param fit_latent: if true, fit the latent models using the dataset, otherwise assume they have already been fit
        :param epochs: number of epochs to use in training of final model
        :return: None
        """
        if fit_latent:
            log.debug("fitting latent models...")
            self.user_latent_model.fit(dataset)
            self.product_latent_model.fit(dataset)

        # process dataset through latent and feature models
        log.debug("encoding dataset using latent and feature models...")
        user_latent = self.user_latent_model.transform(dataset)
        product_latent = self.product_latent_model.transform(dataset)
        features = self.feature_model.transform(dataset)

        X_train = np.hstack([user_latent, product_latent, features])
        log.debug(f"created X_train of shape {X_train.shape}")
        self.input_dim = X_train.shape[1]

        # build and train final_model neural network
        log.debug("building final model...")
        input_layer = Input(shape=(self.input_dim,))
        layers = [input_layer]
        for h_dim in self.h_dims:
            layers.append(Dense(h_dim, activation='relu')(layers[-1]))
        output_layer = Dense(self.output_dim, activation='sigmoid')(layers[-1])

        self.final_model = Model(input_layer, output_layer)
        self.final_model.compile(optimizer='adadelta', loss=tf.keras.losses.BinaryCrossentropy())

        log.debug("fitting final model...")
        self.final_model.fit(X_train, dataset.labels, epochs=epochs, verbose=(log.getEffectiveLevel() == logging.DEBUG))


    def predict(self, dataset):
        """
        predict labels for given dataset

        :param dataset: test dataset
        :return: numpy array of predictions (also saved in self.preds)
        """
        # process dataset through latent and feature models
        log.debug("encoding dataset using latent and feature models...")
        user_latent = self.user_latent_model.transform(dataset)
        product_latent = self.product_latent_model.transform(dataset)
        features = self.feature_model.transform(dataset)

        X_test = np.hstack([user_latent, product_latent, features])
        log.debug(f"created X_test of shape {X_test.shape}")

        log.debug("predicting using final model...")
        self.preds = self.final_model.predict(X_test).squeeze()

        return self.preds


class TopModel(RecModel):
    """
    Main topological recommendation model

    details TBA
    """
    def __init__(self, user_latent_model=None, product_latent_model=None, feature_model=None,
                 h_dims = (100, 50), output_dim=1, n_components=5, NRNN=3, bypass=True):
        super().__init__()
        log.debug("creating main topological model...")
        if user_latent_model is None:
            user_latent_model = EmptyLatentModel()
        if product_latent_model is None:
            product_latent_model = EmptyLatentModel()
        if feature_model is None:
            feature_model = EmptyFeatureModel()
        self.user_latent_model, self.product_latent_model, self.feature_model = user_latent_model, product_latent_model, feature_model

        self.input_dim, self.h_dims, self.output_dim = None, h_dims, output_dim
        self.n_components, self.NRNN, self.bypass = n_components, NRNN, bypass
        self.mapper, self.final_model = MapperClassifier(), None
        self.X_map = None

    def fit(self, dataset, fit_latent=True, epochs=50, retrain_mapper=True):
        """
        Fits model to training dataset

        :param dataset: dataset used for training
        :param fit_latent: if true, fit the latent models using the dataset, otherwise assume they have already been fit
        :param epochs: number of epochs to use in training of final model
        :return: None
        """
        if fit_latent:
            log.debug("fitting latent models...")
            self.user_latent_model.fit(dataset)
            self.product_latent_model.fit(dataset)

        # process dataset through latent and feature models
        log.debug("encoding dataset using latent and feature models...")
        user_latent = self.user_latent_model.transform(dataset)
        product_latent = self.product_latent_model.transform(dataset)
        features = self.feature_model.transform(dataset)
        X_train = np.hstack([user_latent, product_latent, features])
        log.debug(f"created X_train of shape {X_train.shape}")

        # send X_train through mapper-classifier
        log.debug("encoding features with mapper-classifier...")
        if retrain_mapper or self.X_map is None:
            self.X_map = self.mapper.fit_transform(X_train)
        log.debug(f"created mapper encoding of size {self.X_map.shape[1]}")

        # if bypass, include original features alongside encoded features
        if self.bypass:
            self.X_map = np.hstack([X_train, self.X_map])
        log.debug(f"created mapper encoding of size {self.X_map.shape[1]}")

        self.input_dim = self.X_map.shape[1]

        # build and train final_model neural network
        log.debug("building final model...")
        input_layer = Input(shape=(self.input_dim,))
        layers = [input_layer]
        for h_dim in self.h_dims:
            layers.append(Dense(h_dim, activation='relu')(layers[-1]))
        output_layer = Dense(self.output_dim, activation='sigmoid')(layers[-1])

        self.final_model = Model(input_layer, output_layer)
        self.final_model.compile(optimizer='adadelta', loss=tf.keras.losses.BinaryCrossentropy())

        log.debug("fitting final model...")
        self.final_model.fit(self.X_map, dataset.labels, epochs=epochs, verbose=(log.getEffectiveLevel() == logging.DEBUG))


    def predict(self, dataset):
        """
        predict labels for given dataset

        :param dataset: test dataset
        :return: numpy array of predictions (also saved in self.preds)
        """
        # process dataset through latent and feature models
        log.debug("encoding dataset using latent and feature models...")
        user_latent = self.user_latent_model.transform(dataset)
        product_latent = self.product_latent_model.transform(dataset)
        features = self.feature_model.transform(dataset)

        X_test = np.hstack([user_latent, product_latent, features])
        log.debug(f"created X_test of shape {X_test.shape}")

        log.debug("projecting test features using mapper classifier...")
        X_test_map = self.mapper.transform(X_test)

        if self.bypass:
            X_test_map = np.hstack([X_test, X_test_map])

        log.debug("predicting using final model...")
        self.preds = self.final_model.predict(X_test_map).squeeze()

        return self.preds



# TESTING

def run_tests(IC_DATA_DIR):
    from lib.process_data import instacart_process
    from lib.data_class import DataSet
    from models.latent_models import UserModel, ProductModel
    from models.feature_models import MainFeatureModel

    # set random seed for consistent tests
    np.random.seed(42)

    # load data from instacart csv files (values below use testing directory)
    order_data, product_data = instacart_process(data_dir=IC_DATA_DIR)

    # create dataset
    ic_dataset = DataSet(order_df=order_data, product_df=product_data)

    # check dataframes created correctly
    assert ic_dataset.order_df.shape == (31032, 4)
    assert ic_dataset.product_df.shape == (6126, 5)

    # perform train-test split
    train_dataset, test_dataset = ic_dataset.train_test_split()
    assert train_dataset.order_df.shape == (24939, 4)
    assert test_dataset.order_df.shape == (6093, 4)

    # create user latent model, fit and transform
    user_latent = UserModel()
    user_latent.fit(train_dataset, epochs=2)
    assert user_latent.transform(train_dataset).shape == (9447, 32)

    # create product latent model, fit and transform
    product_latent = ProductModel()
    product_latent.fit(train_dataset)
    assert product_latent.transform(train_dataset).shape == (9447, 10)

    # create feature model
    feature_model = MainFeatureModel()
    X = feature_model.transform(train_dataset)
    assert X.shape == (9447, 8)

    # fit non-top model to train_dataset
    model = NonTopModel(user_latent_model=user_latent, product_latent_model=product_latent, feature_model=feature_model)
    model.fit(train_dataset, fit_latent=False, epochs=2)
    assert model.input_dim == 50

    # predict on test_dataset
    model.predict(test_dataset)
    assert model.preds.shape == (2767,)
    assert model.preds.shape == test_dataset.labels.shape

    # fit top model to train_dataset
    model = TopModel(user_latent_model=user_latent, product_latent_model=product_latent, feature_model=feature_model)
    model.fit(train_dataset, fit_latent=False, epochs=2)
    print(model.input_dim)

    # predict on test_dataset
    model.predict(test_dataset)
    assert model.preds.shape == (2767,)
    assert model.preds.shape == test_dataset.labels.shape

    log.info("main_models tests passed!")


if __name__ == '__main__':
    IC_DATA_DIR = '../data/instacart_2017_05_01_testing/'
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    run_tests(IC_DATA_DIR)
